//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2023 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#include <stdafx.h>
#include <vd2/system/math.h>
#include <vd2/system/vdalloc.h>
#include <at/atcore/fft.h>
#include "test.h"

AT_DEFINE_TEST(Core_FFT) {
	vdautoptr<ATFFT<32>> fft32(new ATFFT<32>);
	vdautoptr<ATFFT<64>> fft64(new ATFFT<64>);
	vdautoptr<ATFFT<4096>> fft4k(new ATFFT<4096>);

	struct xdat {
		alignas(64) float x[4096];
		alignas(64) float y[4096];
		alignas(64) float ref[4096];
		alignas(64) float costab[4096];
	};
	vdautoptr<xdat> dat0(new xdat);
	xdat *dat = dat0;

	for(int i=0; i<32; ++i) {
		std::fill(dat->x, dat->x + 32, 0.0f);
		dat->x[i] = 1;

		// compute reference output (slowly)
		float ref[32];
		const float div = -nsVDMath::kfTwoPi / 32.0f;
		for(int j=0; j<=16; ++j) {
			float rsum = 0;
			float isum = 0;

			for(int k=0; k<32; ++k) {
				float w = (div * j) * k;
				rsum += dat->x[k] * cosf(w);
				isum += dat->x[k] * sinf(w);
			}

			if (j == 16)
				ref[1] = rsum;
			else {
				ref[j*2] = rsum;
				ref[j*2+1] = isum;
			}
		}

		fft32->Forward(dat->x);

		bool fail = false;
		for(int j=0; j<32; ++j) {
			if (fabsf(dat->x[j] - ref[j]) > 1e-3f)
				fail = true;
		}

		memcpy(dat->y, dat->x, sizeof dat->y);
		fft32->Inverse(dat->y);

		for(int j=0; j<32; ++j) {
			if (i == j) {
				if (fabsf(dat->y[j] - 16.0f) > 1e-3f)
					fail = true;
			} else {
				if (fabsf(dat->y[j]) > 2e-3f)
					fail = true;
			}
		}

		if (fail) {
			printf("FFT32 fail\n");

			for(int j=0; j<32; ++j)
				printf(" %4.1f", dat->x[j]);

			putchar('\n');

			for(int j=0; j<32; ++j)
				printf(" %4.1f", dat->y[j]);

			putchar('\n');
			return 1;
		}
	}

	AT_TEST_TRACE("FFT32 pass");

	for(int i=0; i<4096; ++i) {
		std::fill(dat->x, dat->x + 64, 0.0f);
		dat->x[i] = 1;

		fft64->Forward(dat->x);
		fft64->Inverse(dat->x);

		bool fail = false;
		for(int j=0; j<64; ++j) {
			if (i == j) {
				if (fabsf(dat->x[j] - 32.0f) > 1e-3f)
					fail = true;
			} else {
				if (fabsf(dat->x[j]) > 2e-3f)
					fail = true;
			}
		}

		if (fail) {
			printf("FFT64 fail\n");
			return 1;
		}
	}

	AT_TEST_TRACE("FFT64 pass");

	// we only test a subset of the 4K positions for speed reasons
	for(int i=0; i<4096; i += 37) {
		std::fill(dat->x, dat->x + 4096, 0.0f);
		dat->x[i] = 1;

		// spot check only a single reference bucket, since it's very slow
		if (i == 37*15) {
			const float div = -nsVDMath::kfTwoPi / 4096.0f;
			for(int j=0; j<4096; ++j) {
				dat->costab[j] = cosf(div * j);
			}

			for(int j=0; j<=2048; ++j) {
				float rsum = 0;
				float isum = 0;
				 
				for(int k=0; k<4096; ++k) {
					rsum += dat->x[k] * dat->costab[(j*k) & 4095];
					isum += dat->x[k] * dat->costab[(j*k + 1024) & 4095];
				}

				if (j == 2048)
					dat->ref[1] = rsum;
				else {
					dat->ref[j*2] = rsum;
					dat->ref[j*2+1] = isum;
				}
			}
		}

		fft4k->Forward(dat->x);

		if (i == 37*15) {
			for(int j=0; j<4096; ++j) {
				if (fabsf(dat->x[j] - dat->ref[j]) > 1e-3f) {
					printf("FFT4K forward fail\n");
					return 1;
				}
			}
		}

		fft4k->Inverse(dat->x);

		bool fail = false;
		for(int j=0; j<4096; ++j) {
			if (i == j) {
				if (fabsf(dat->x[j] - 2048.0f) > 1e-3f)
					fail = true;
			} else {
				if (fabsf(dat->x[j]) > 2e-3f)
					fail = true;
			}
		}

		if (fail) {
			printf("FFT4K fail\n");
			return 1;
		}
	}

	AT_TEST_TRACE("FFT4K pass");

	return 0;
}

AT_DEFINE_TEST(Core_IMDCT) {
	float x[64];
	float costab[512];

	for(int i=0; i<512; ++i) {
		costab[i] = cosf(-nsVDMath::kfTwoPi / 512 * i);
	}

	ATFFTAllocator alloc;
	ATIMDCT imdct;
	imdct.Reserve(alloc, 64);
	alloc.Finalize();
	imdct.Bind(alloc);
	for(int i=0; i<64; ++i) {
		for(float& v : x)
			v = 0;

		x[i] = 1.0f;

		float ref[64];
		for(int j=0; j<32; ++j) {
			float sum = 0;

			for(int k = 0; k<64; ++k)
				sum += x[k] * costab[((j*2 + (1 + 64 + 64*2)) * (k*2 + 1)) & 511];

			ref[j] = sum;
		}

		for(int j=0; j<32; ++j) {
			float sum = 0;

			for(int k = 0; k<64; ++k)
				sum += x[k] * costab[((j*2 + (1 + 64 + 32*2)) * (k*2 + 1)) & 511];

			ref[j + 32] = sum;
		}

		imdct.Transform(x);

		float me = 0;
		float msqe = 0;
		float maxe = 0;
		for(int j=0; j<64; ++j) {
			const float e = x[j] - ref[j];
			me += e;
			msqe += e*e;
			maxe = std::max<float>(maxe, fabsf(e));
		}

		me /= 64.0f;
		msqe /= 64.0f;

		AT_TEST_TRACEF("%2u) mean error = %g, mean squared error = %g, max error = %g", i, me, msqe, maxe);

		AT_TEST_ASSERT(fabsf(me) < 1e-3f && msqe < 1e-3f);
	}

	return 0;
}
