//	Altirra - Atari 800/800XL emulator
//	Copyright (C) 2009 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <stdafx.h>
#include <vd2/system/error.h>
#include <vd2/system/strutil.h>
#include <at/atcore/atascii.h>
#include <at/atcore/cio.h>
#include <at/atcore/devicecio.h>
#include <at/atcore/deviceimpl.h>
#include <at/atcore/deviceparent.h>
#include <at/atcore/deviceparentimpl.h>
#include <at/atcore/deviceprinter.h>
#include <at/atcore/devicesio.h>
#include "kerneldb.h"

class ATDevicePrinterBase : public ATDevice, public IATDevicePrinterPort, public IATDeviceCIO, public IATDeviceSIO, public IATDeviceParent {
public:
	ATDevicePrinterBase(bool lineBuffered, bool sidewaysSupported);
	virtual ~ATDevicePrinterBase();

	void *AsInterface(uint32 id) override;

	void Init() override;
	void Shutdown() override;
	void WarmReset() override;
	void ColdReset() override;

public:	// IATDeviceCIO
	void InitCIO(IATDeviceCIOManager *mgr) override;
	void GetCIODevices(char *buf, size_t len) const override;
	sint32 OnCIOOpen(int channel, uint8 deviceNo, uint8 aux1, uint8 aux2, const uint8 *filename) override;
	sint32 OnCIOClose(int channel, uint8 deviceNo) override;
	sint32 OnCIOGetBytes(int channel, uint8 deviceNo, void *buf, uint32 len, uint32& actual) override;
	sint32 OnCIOPutBytes(int channel, uint8 deviceNo, const void *buf, uint32 len, uint32& actual) override;
	sint32 OnCIOGetStatus(int channel, uint8 deviceNo, uint8 statusbuf[4]) override;
	sint32 OnCIOSpecial(int channel, uint8 deviceNo, uint8 cmd, uint16 bufadr, uint16 buflen, uint8 aux[6]) override;
	void OnCIOAbortAsync() override;

public:	// IATDevicePrinterPort
	void SetPrinterDefaultOutput(IATPrinterOutput *output) override;

public:	// IATDeviceSIO
	void InitSIO(IATDeviceSIOManager *mgr) override;
	CmdResponse OnSerialBeginCommand(const ATDeviceSIOCommand& cmd) override;
	void OnSerialAbortCommand() override;
	void OnSerialReceiveComplete(uint32 id, const void *data, uint32 len, bool checksumOK) override;
	void OnSerialFence(uint32 id) override;
	CmdResponse OnSerialAccelCommand(const ATDeviceSIORequest& request) override;

public:	// IATDeviceParent
	IATDeviceBus *GetDeviceBus(uint32 index);

protected:
	void GetStatusFrame(uint8 frame[4]);
	virtual void GetStatusFrameInternal(uint8 frame[4]) = 0;

	void FlushCIOBuffer(bool sideways);

	void HandleFrame(const uint8 *data, uint32 len);
	virtual void HandleFrameInternal(IATPrinterOutput& output, uint8 orientation, uint8 *buf, uint32 len) = 0;

	IATDeviceCIOManager *mpCIOMgr = nullptr;
	IATDeviceSIOManager *mpSIOMgr = nullptr;
	IATPrinterOutput *mpOutput = nullptr;

	const bool mbLineBuffered;
	const bool mbSidewaysSupported;

	uint8		mLastOrientationByte = 0xFF;

	uint8		mCIOBuffer[40] {};
	uint8		mCIOBufferIndex = 0;

	ATDeviceBusSingleChild mParallelBus;
};

ATDevicePrinterBase::ATDevicePrinterBase(bool lineBuffered, bool sidewaysSupported)
	: mbLineBuffered(lineBuffered)
	, mbSidewaysSupported(sidewaysSupported)
{
}

ATDevicePrinterBase::~ATDevicePrinterBase() {
}

void *ATDevicePrinterBase::AsInterface(uint32 id) {
	switch(id) {
		case IATDevicePrinterPort::kTypeID:
			return static_cast<IATDevicePrinterPort *>(this);
		case IATDeviceSIO::kTypeID:
			return static_cast<IATDeviceSIO *>(this);
		case IATDeviceCIO::kTypeID:
			return static_cast<IATDeviceCIO *>(this);
		case IATDeviceParent::kTypeID:
			return static_cast<IATDeviceParent *>(this);
	}

	return ATDevice::AsInterface(id);
}

void ATDevicePrinterBase::Init() {
	mParallelBus.Init(this, 0, IATPrinterOutput::kTypeID, "parallel", L"Printer Output", "parport");

	ColdReset();
}

void ATDevicePrinterBase::Shutdown() {
	mParallelBus.Shutdown();

	if (mpCIOMgr) {
		mpCIOMgr->RemoveCIODevice(this);
		mpCIOMgr = nullptr;
	}

	if (mpSIOMgr) {
		mpSIOMgr->RemoveDevice(this);
		mpSIOMgr = nullptr;
	}

	mpOutput = nullptr;
}

void ATDevicePrinterBase::WarmReset() {
	ColdReset();
}

void ATDevicePrinterBase::ColdReset() {
	mCIOBufferIndex = 0;
}

void ATDevicePrinterBase::InitCIO(IATDeviceCIOManager *mgr) {
	mpCIOMgr = mgr;
	mpCIOMgr->AddCIODevice(this);
}

void ATDevicePrinterBase::GetCIODevices(char *buf, size_t len) const {
	vdstrlcpy(buf, "P", len);
}

sint32 ATDevicePrinterBase::OnCIOOpen(int channel, uint8 deviceNo, uint8 aux1, uint8 aux2, const uint8 *filename) {
	return kATCIOStat_NotSupported;
}

sint32 ATDevicePrinterBase::OnCIOClose(int channel, uint8 deviceNo) {
	if (mCIOBufferIndex)
		FlushCIOBuffer(mpCIOMgr->ReadByte(ATKernelSymbols::ICAX2Z) == 0x53);

	return kATCIOStat_Success;
}

sint32 ATDevicePrinterBase::OnCIOGetBytes(int channel, uint8 deviceNo, void *buf, uint32 len, uint32& actual) {
	return kATCIOStat_ReadOnly;
}

sint32 ATDevicePrinterBase::OnCIOPutBytes(int channel, uint8 deviceNo, const void *buf, uint32 len, uint32& actual) {
	actual = len;

	// read orientation byte and check if it is sideways; everything else treated as normal
	const uint8 cioOrientation = mpCIOMgr->ReadByte(ATKernelSymbols::ICAX2Z);
	const bool sideways = (cioOrientation == 0x53);
	const uint32 width = sideways ? 29 : 40;

	// buffer bytes and flush after EOL or buffer full
	const uint8 *src = (const uint8 *)buf;
	while(len) {
		if (mCIOBufferIndex >= width)
			FlushCIOBuffer(sideways);

		const uint8 ch = *src++;
		mCIOBuffer[mCIOBufferIndex++] = ch;

		if (ch == 0x9B)
			FlushCIOBuffer(sideways);

		--len;
	}

	if (!mbLineBuffered)
		FlushCIOBuffer(sideways);

	return kATCIOStat_Success;
}

sint32 ATDevicePrinterBase::OnCIOGetStatus(int channel, uint8 deviceNo, uint8 statusbuf[4]) {
	GetStatusFrame(statusbuf);
	return kATCIOStat_Success;
}

sint32 ATDevicePrinterBase::OnCIOSpecial(int channel, uint8 deviceNo, uint8 cmd, uint16 bufadr, uint16 buflen, uint8 aux[6]) {
	return kATCIOStat_NotSupported;
}

void ATDevicePrinterBase::OnCIOAbortAsync() {
}

void ATDevicePrinterBase::SetPrinterDefaultOutput(IATPrinterOutput *output) {
	mpOutput = output;
}

void ATDevicePrinterBase::InitSIO(IATDeviceSIOManager *mgr) {
	mpSIOMgr = mgr;
	mpSIOMgr->AddDevice(this);
}

IATDeviceSIO::CmdResponse ATDevicePrinterBase::OnSerialBeginCommand(const ATDeviceSIOCommand& cmd) {
	if (!cmd.mbStandardRate)
		return kCmdResponse_NotHandled;

	if (cmd.mDevice != 0x40)
		return kCmdResponse_NotHandled;

	if (cmd.mCommand == 'W') {
		mLastOrientationByte = cmd.mAUX[0];

		if (cmd.mAUX[0] == 'S' && mbSidewaysSupported) {
			mpSIOMgr->BeginCommand();
			mpSIOMgr->SendACK();
			mpSIOMgr->ReceiveData(0, 29, true);
			mpSIOMgr->SendComplete();
		} else if (cmd.mAUX[0] == 'N') {
			mpSIOMgr->BeginCommand();
			mpSIOMgr->SendACK();
			mpSIOMgr->ReceiveData(0, 40, true);
			mpSIOMgr->SendComplete();
		} else {
			return kCmdResponse_Fail_NAK;
		}

		mpSIOMgr->EndCommand();

		return kCmdResponse_Start;
	} else if (cmd.mCommand == 'S') {
		uint8 statusData[4] {};

		GetStatusFrame(statusData);

		mpSIOMgr->HandleCommand(statusData, 4, true);
		return kCmdResponse_Start;
	}

	return kCmdResponse_NotHandled;
}

void ATDevicePrinterBase::OnSerialAbortCommand() {
}

void ATDevicePrinterBase::OnSerialReceiveComplete(uint32 id, const void *data, uint32 len, bool checksumOK) {
	HandleFrame((const uint8 *)data, len);
}

void ATDevicePrinterBase::OnSerialFence(uint32 id) {
}

IATDeviceSIO::CmdResponse ATDevicePrinterBase::OnSerialAccelCommand(const ATDeviceSIORequest& request) {
	return OnSerialBeginCommand(request);
}

IATDeviceBus *ATDevicePrinterBase::GetDeviceBus(uint32 index) {
	return index ? nullptr : &mParallelBus;
}

void ATDevicePrinterBase::GetStatusFrame(uint8 frame[4]) {
	frame[0] = 0;
	frame[1] = 0;
	frame[2] = 16;
	frame[3] = 0;

	GetStatusFrameInternal(frame);
}

void ATDevicePrinterBase::FlushCIOBuffer(bool sideways) {
	if (!mCIOBufferIndex)
		return;

	uint32 width = mCIOBufferIndex;
	
	if (mbLineBuffered) {
		width = sideways ? 29 : 40;

		// pad buffer with EOLs if we have a short line
		if (mCIOBufferIndex < width)
			memset(&mCIOBuffer[mCIOBufferIndex], 0x9B, width - mCIOBufferIndex);
	}

	// set last orientation byte N/S, since we are short-circuiting an actual SIO frame; note
	// that this isn't just ICAX2Z as unrecognized values need to be converted to Normal
	mLastOrientationByte = sideways ? 0x53 : 0x4E;

	// send the (whole) buffer
	HandleFrame(mCIOBuffer, width);

	// reset buffer index
	mCIOBufferIndex = 0;
}

void ATDevicePrinterBase::HandleFrame(const uint8 *data, uint32 len) {
	const uint8 *src = (const uint8 *)data;

	if (len > 40)
		len = 40;

	// Ensure that the frame is EOL terminated. The additional byte is also guaranteed
	// to be available to HandleFrameInternal().
	uint8 buf[41];
	memcpy(buf, src, len);
	buf[len] = 0x9B;
	
	auto *printer = mParallelBus.GetChild<IATPrinterOutput>();
	if (!printer) {
		printer = mpOutput;
		if (!printer)
			return;
	}

	HandleFrameInternal(*printer, mLastOrientationByte, buf, len);
}

///////////////////////////////////////////////////////////////////////////////

class ATDevicePrinter final : public ATDevicePrinterBase {
public:
	ATDevicePrinter();
	~ATDevicePrinter();

	void GetDeviceInfo(ATDeviceInfo& info) override;

	void GetStatusFrameInternal(uint8 frame[4]) override;
	void HandleFrameInternal(IATPrinterOutput& output, uint8 orientation, uint8 *buf, uint32 len) override;
};

void ATCreateDevicePrinter(const ATPropertySet& pset, IATDevice **dev) {
	vdrefptr<ATDevicePrinter> p(new ATDevicePrinter);

	*dev = p.release();
}

extern const ATDeviceDefinition g_ATDeviceDefPrinter = { "printer", nullptr, L"Printer (P:)", ATCreateDevicePrinter };

ATDevicePrinter::ATDevicePrinter()
	: ATDevicePrinterBase(false, true)
{
	SetSaveStateAgnostic();
}

ATDevicePrinter::~ATDevicePrinter() {
}

void ATDevicePrinter::GetDeviceInfo(ATDeviceInfo& info) {
	info.mpDef = &g_ATDeviceDefPrinter;
}

void ATDevicePrinter::GetStatusFrameInternal(uint8 frame[4]) {
}

void ATDevicePrinter::HandleFrameInternal(IATPrinterOutput& output, uint8 orientation, uint8 *buf, uint32 len) {
	// look for an EOL, and if there is one, truncate at that point and set eol flag
	bool eol = false;

	for(uint32 i=0; i<len; ++i) {
		if (buf[i] == 0x9B) {
			eol = true;
			len = i;
		}
	}

	// if there is an EOL, then trim off spaces at end of the line, and add EOL at the end
	// of the buffer (we are guaranteed that an extra buffer space is available)
	if (eol) {
		while(len && buf[len - 1] == 0x20)
			--len;

		buf[len++] = 0x9B;
	}

	if (!output.WantUnicode()) {
		output.WriteRaw(buf, len);
		return;
	}

	wchar_t unibuf[41];
	for(uint32 i=0; i<len; ++i) {
		uint8 c = *buf++;

		unibuf[i] = c == 0x9B ? '\n' : (wchar_t)c;
	}

	output.WriteUnicode(unibuf, len);
}

///////////////////////////////////////////////////////////////////////////////

class ATDevicePrinter820 final : public ATDevicePrinterBase {
public:
	ATDevicePrinter820();
	~ATDevicePrinter820();

	void GetDeviceInfo(ATDeviceInfo& info) override;

	void GetStatusFrameInternal(uint8 frame[4]) override;
	void HandleFrameInternal(IATPrinterOutput& output, uint8 orientation, uint8 *buf, uint32 len) override;
};

void ATCreateDevicePrinter820(const ATPropertySet& pset, IATDevice **dev) {
	vdrefptr<ATDevicePrinter820> p(new ATDevicePrinter820);

	*dev = p.release();
}

extern const ATDeviceDefinition g_ATDeviceDefPrinter820 = { "820", nullptr, L"820 Printer", ATCreateDevicePrinter820 };

ATDevicePrinter820::ATDevicePrinter820()
	: ATDevicePrinterBase(true, true)
{
	SetSaveStateAgnostic();
}

ATDevicePrinter820::~ATDevicePrinter820() {
}

void ATDevicePrinter820::GetDeviceInfo(ATDeviceInfo& info) {
	info.mpDef = &g_ATDeviceDefPrinter820;
}

void ATDevicePrinter820::GetStatusFrameInternal(uint8 frame[4]) {
	// The 820 sets the 'intelligent controller' bit 7 on device status.
	frame[0] |= 0x80;

	// The orientation byte from the last Write command (AUX1) is reported as second status byte
	frame[1] = mLastOrientationByte;

	// 820 timeout is 6*64 vblanks
	frame[2] = 6;

	// unused byte is $00
	frame[3] = 0;
}

void ATDevicePrinter820::HandleFrameInternal(IATPrinterOutput& output, uint8 orientation, uint8 *buf, uint32 len) {
	// check for sideways
	const bool sideways = (orientation == 0x53);
	if (sideways) {
		// sideways: mask off bit 7, convert $60-7F to 40-5F, then convert anything not in
		// $30-5F afterward to space, and write backwards
		for(uint32 i = 0; i < len; ++i) {
			buf[i] &= 0x7F;

			if (buf[i] < 0x20 || buf[i] > 0x7F)
				buf[i] = 0x20;
			else if (buf[i] >= 0x60)
				buf[i] -= 0x20;
		}

		std::reverse(buf, buf + len);
	} else {
		// normal: mask off bit 7, and convert anything not in $20-7E afterward to space
		for(uint32 i = 0; i < len; ++i) {
			buf[i] &= 0x7F;

			if (buf[i] < 0x20 || buf[i] > 0x7E)
				buf[i] = 0x20;
		}
	}

	// trim off any spaces at the end, just to make the printer output nicer
	while(len && buf[len - 1] == 0x20)
		--len;

	// put in a new EOL; buf[] is len+1 so this is allowed
	buf[len++] = 0x9B;

	// write out line
	if (!output.WantUnicode()) {
		output.WriteRaw(buf, len);
		return;
	}

	// 820 character set is actually closer to ASCII than ATASCII. Normal mode is
	// entirely ASCII; sideways has up and left arrow for $5E/5F.
	wchar_t unibuf[41];

	if (sideways) {
		for(uint32 i=0; i<len; ++i) {
			wchar_t ch = buf[i];

			switch(ch) {
				case 0x5E:
					ch = L'\u2191';		// U+2191 Upwards arrow
					break;
				case 0x5F:
					ch = L'\u2190';		// U+2190 Leftwards arrow
					break;
				case 0x9B:
					ch = L'\n';
					break;
			}

			unibuf[i] = ch;
		}
	} else {
		for(uint32 i=0; i<len; ++i) {
			const uint8 c = buf[i];

			unibuf[i] = c == 0x9B ? L'\n' : c;
		}
	}

	output.WriteUnicode(unibuf, len);
}

///////////////////////////////////////////////////////////////////////////////

class ATDevicePrinter1025 final : public ATDevicePrinterBase {
public:
	ATDevicePrinter1025();
	~ATDevicePrinter1025();

	void GetDeviceInfo(ATDeviceInfo& info) override;
	void ColdReset() override;

	void GetStatusFrameInternal(uint8 frame[4]) override;
	void HandleFrameInternal(IATPrinterOutput& output, uint8 orientation, uint8 *buf, uint32 len) override;

private:
	void FlushLine(IATPrinterOutput& output);
	void UpdateLineLength();
	void ResetState();

	uint8 mColumn = 0;
	uint8 mLineLength = 0;

	enum class CharDensity : uint8 {
		Cpi5,
		Cpi10,
		Cpi16_5,
	} mCharDensity = CharDensity::Cpi10;

	bool mbEscapeAtStart = false;
	bool mbEscape = false;
	bool mbEuropeanChars = false;
	bool mbDenseLines = false;
	bool mbShortLine = false;

	uint8 mRawLineBuffer[133] {};
	wchar_t mUnicodeLineBuffer[133] {};
};

void ATCreateDevicePrinter1025(const ATPropertySet& pset, IATDevice **dev) {
	vdrefptr<ATDevicePrinter1025> p(new ATDevicePrinter1025);

	*dev = p.release();
}

extern const ATDeviceDefinition g_ATDeviceDefPrinter1025 = { "1025", nullptr, L"1025 Printer", ATCreateDevicePrinter1025 };

ATDevicePrinter1025::ATDevicePrinter1025()
	: ATDevicePrinterBase(false, true)
{
	SetSaveStateAgnostic();
}

ATDevicePrinter1025::~ATDevicePrinter1025() {
}

void ATDevicePrinter1025::GetDeviceInfo(ATDeviceInfo& info) {
	info.mpDef = &g_ATDeviceDefPrinter1025;
}

void ATDevicePrinter1025::ColdReset() {
	ResetState();
}

void ATDevicePrinter1025::GetStatusFrameInternal(uint8 frame[4]) {
	// The 1025 does not return last orientation byte.
	frame[1] = 0;

	// 1025 timeout is 15*64 vblanks
	frame[2] = 15;

	// unused byte is $#0
	frame[3] = 0x30;

	// Rather unusually, the 1025 also resets all of its internal state when processing
	// a Status command. This makes it sensitive to whether LPRINT or PRINT is used,
	// since in the former case BASIC opens and closes P: for each line. This means that
	// settings from ESC codes, like density and European enables, get reset.

	ResetState();
}

void ATDevicePrinter1025::HandleFrameInternal(IATPrinterOutput& output, uint8 orientation, uint8 *buf, uint32 len) {
	// normal: mask off bit 7, and convert 00-1F/60/7B/7D-7F to space
	// european: just mask off bit 7
	//
	// The following escape codes are handled at any time:
	//	ESC CTRL+T		Print 16.5 CPI
	//	ESC CTRL+O		Print 10 CPI
	//	ESC CTRL+N		Print 5 CPI
	//	ESC L			Set long line (80 chars)
	//	ESC S			Set short line (64 chars)
	//	ESC 6			Set to 6 LPI
	//	ESC 8			Set to 8 LPI
	//
	// The following escape codes are only handled at start of line:
	//	ESC CTRL+W		enable european characters
	//	ESC CTRL+X		ignore european characters
	//
	// Regardless of whether the escape is valid, the character is omitted
	// from the line buffer and escape modes are cleared after processing an
	// escape. EOL takes priority over ESC and escape is cleared after processing
	// an EOL.

	for(uint32 i = 0; i < len; ++i) {
		uint8 c = buf[i];

		// check for EOL
		if (c == 0x9B) {
			// flush line buffer
			FlushLine(output);

			// ignore remainder of data frame
			break;
		}

		// check for ESC
		if (c == 0x1B) {
			if (mColumn == 0)
				mbEscapeAtStart = true;
			else
				mbEscape = true;

			continue;
		} else if (mbEscapeAtStart || mbEscape) {
			if (mbEscapeAtStart) {
				switch(c) {
					case 0x14:	// Ctrl+T
						mCharDensity = CharDensity::Cpi16_5;
						break;

					case 0x0F:	// Ctrl+O
						mCharDensity = CharDensity::Cpi10;
						break;

					case 0x0E:	// Ctrl+N
						mCharDensity = CharDensity::Cpi5;
						break;

					case 0x4C:	// L
						mbShortLine = false;
						break;

					case 0x53:	// S
						mbShortLine = true;
						break;

					case 0x36:	// 6
						mbDenseLines = false;
						break;

					case 0x38:	// 8
						mbDenseLines = true;
						break;

					default:
						break;
				}
			}

			switch(c) {
				case 0x17:	// Ctrl+W
					mbEuropeanChars = true;
					break;

				case 0x18:	// Ctrl+X
					mbEuropeanChars = false;
					break;
			}

			mbEscape = false;
			mbEscapeAtStart = false;
			continue;
		}

		// Printable character -- put in line buffer.
		//
		// The 1025's character set is the international ATASCII set. In European
		// mode, the entire 7-bit character set is accessible except for ESC, which
		// cannot be printed (ESC ESC is a no-op). When European characters are
		// disabled, the characters that differ are suppressed -- 00-1F are ignored
		// and 60/7B/7D-7F are blanked.
		//
		// Bit 7 is ignored.

		c &= 0x7F;

		if (!mbEuropeanChars) {
			if (c < 0x20)
				continue;

			if (c == 0x60 || c == 0x7B || c >= 0x7D)
				c = 0x20;
		}

		if (!mColumn)
			UpdateLineLength();

		mRawLineBuffer[mColumn] = c;
		mUnicodeLineBuffer[mColumn] = kATATASCIITables.mATASCIIToUnicode[1][c];
		if (++mColumn >= mLineLength)
			FlushLine(output);
	}
}

void ATDevicePrinter1025::FlushLine(IATPrinterOutput& output) {
	// trim off any spaces at the end, just to make the printer output nicer
	while(mColumn && mRawLineBuffer[mColumn - 1] == 0x20)
		--mColumn;

	mRawLineBuffer[mColumn] = 0x9B;
	mUnicodeLineBuffer[mColumn] = L'\n';

	if (output.WantUnicode())
		output.WriteUnicode(mUnicodeLineBuffer, mColumn + 1);
	else
		output.WriteRaw(mRawLineBuffer, mColumn + 1);

	mColumn = 0;
	mbEscape = false;
	mbEscapeAtStart = false;
}

void ATDevicePrinter1025::UpdateLineLength() {
	switch(mCharDensity) {
		case CharDensity::Cpi5:
			mLineLength = mbShortLine ? 32 : 40;
			break;

		case CharDensity::Cpi10:
		default:
			mLineLength = mbShortLine ? 64 : 80;
			break;

		case CharDensity::Cpi16_5:
			mLineLength = mbShortLine ? 105 : 132;
			break;
	}
}

void ATDevicePrinter1025::ResetState() {
	// According to the Atari 1025 Printer Field Service Manual, page 3-3, the default power-on
	// state is: 10 CPI, 6 LPI, 80 chars/line, European chars disabled.
	mbEscape = false;
	mbEscapeAtStart = false;
	mbEuropeanChars = false;
	mCharDensity = CharDensity::Cpi10;
	mbDenseLines = false;
	mbShortLine = false;
}
